@@ -20,6 +20,12 @@ NVL <- function(x, val) {
   stop("typeof(x) == ", typeof(x), " is not supported by NVL")
 }
 
+# List of classification and ranking objectives
+.CLASSIFICATION_OBJECTIVES <- function() {
+  return(c('binary:logistic', 'binary:logitraw', 'binary:hinge', 'multi:softmax',
+           'multi:softprob', 'rank:pairwise', 'rank:ndcg', 'rank:map'))
+}
+
 
 #
 # Low-level functions for boosting --------------------------------------------
@@ -69,23 +75,23 @@ check.booster.params <- function(params, ...) {
 
   if (!is.null(params[['monotone_constraints']]) &&
       typeof(params[['monotone_constraints']]) != "character") {
-        vec2str = paste(params[['monotone_constraints']], collapse = ',')
-        vec2str = paste0('(', vec2str, ')')
-        params[['monotone_constraints']] = vec2str
+        vec2str <- paste(params[['monotone_constraints']], collapse = ',')
+        vec2str <- paste0('(', vec2str, ')')
+        params[['monotone_constraints']] <- vec2str
   }
 
   # interaction constraints parser (convert from list of column indices to string)
   if (!is.null(params[['interaction_constraints']]) &&
       typeof(params[['interaction_constraints']]) != "character"){
     # check input class
-    if (!identical(class(params[['interaction_constraints']]),'list')) stop('interaction_constraints should be class list')
-    if (!all(unique(sapply(params[['interaction_constraints']], class)) %in% c('numeric','integer'))) {
+    if (!identical(class(params[['interaction_constraints']]), 'list')) stop('interaction_constraints should be class list')
+    if (!all(unique(sapply(params[['interaction_constraints']], class)) %in% c('numeric', 'integer'))) {
       stop('interaction_constraints should be a list of numeric/integer vectors')
     }
 
     # recast parameter as string
-    interaction_constraints <- sapply(params[['interaction_constraints']], function(x) paste0('[', paste(x, collapse=','), ']'))
-    params[['interaction_constraints']] <- paste0('[', paste(interaction_constraints, collapse=','), ']')
+    interaction_constraints <- sapply(params[['interaction_constraints']], function(x) paste0('[', paste(x, collapse = ','), ']'))
+    params[['interaction_constraints']] <- paste0('[', paste(interaction_constraints, collapse = ','), ']')
   }
   return(params)
 }
@@ -145,7 +151,8 @@ xgb.iter.update <- function(booster_handle, dtrain, iter, obj = NULL) {
   if (is.null(obj)) {
     .Call(XGBoosterUpdateOneIter_R, booster_handle, as.integer(iter), dtrain)
   } else {
-    pred <- predict(booster_handle, dtrain, outputmargin = TRUE, training = TRUE)
+    pred <- predict(booster_handle, dtrain, outputmargin = TRUE, training = TRUE,
+                    ntreelimit = 0)
     gpair <- obj(pred, dtrain)
     .Call(XGBoosterBoostOneIter_R, booster_handle, dtrain, gpair$grad, gpair$hess)
   }
@@ -166,13 +173,12 @@ xgb.iter.eval <- function(booster_handle, watchlist, iter, feval = NULL) {
   evnames <- names(watchlist)
   if (is.null(feval)) {
     msg <- .Call(XGBoosterEvalOneIter_R, booster_handle, as.integer(iter), watchlist, as.list(evnames))
-    msg <- stri_split_regex(msg, '(\\s+|:|\\s+)')[[1]][-1]
-    res <- as.numeric(msg[c(FALSE,TRUE)]) # even indices are the values
-    names(res) <- msg[c(TRUE,FALSE)]      # odds are the names
+    mat <- matrix(strsplit(msg, '\\s+|:')[[1]][-1], nrow = 2)
+    res <- structure(as.numeric(mat[2, ]), names = mat[1, ])
   } else {
     res <- sapply(seq_along(watchlist), function(j) {
       w <- watchlist[[j]]
-      preds <- predict(booster_handle, w) # predict using all trees
+      preds <- predict(booster_handle, w, outputmargin = TRUE, ntreelimit = 0) # predict using all trees
       eval_res <- feval(preds, w)
       out <- eval_res$value
       names(out) <- paste0(evnames[j], "-", eval_res$metric)
@@ -187,13 +193,23 @@ xgb.iter.eval <- function(booster_handle, watchlist, iter, feval = NULL) {
 # Helper functions for cross validation ---------------------------------------
 #
 
+# Possibly convert the labels into factors, depending on the objective.
+# The labels are converted into factors only when the given objective refers to the classification
+# or ranking tasks.
+convert.labels <- function(labels, objective_name) {
+  if (objective_name %in% .CLASSIFICATION_OBJECTIVES()) {
+    return(as.factor(labels))
+  } else {
+    return(labels)
+  }
+}
+
 # Generates random (stratified if needed) CV folds
 generate.cv.folds <- function(nfold, nrows, stratified, label, params) {
 
   # cannot do it for rank
-  if (exists('objective', where = params) &&
-      is.character(params$objective) &&
-      strtrim(params$objective, 5) == 'rank:') {
+  objective <- params$objective
+  if (is.character(objective) && strtrim(objective, 5) == 'rank:') {
     stop("\n\tAutomatic generation of CV-folds is not implemented for ranking!\n",
          "\tConsider providing pre-computed CV-folds through the 'folds=' parameter.\n")
   }
@@ -206,19 +222,16 @@ generate.cv.folds <- function(nfold, nrows, stratified, label, params) {
     #  - For classification, need to convert y labels to factor before making the folds,
     #    and then do stratification by factor levels.
     #  - For regression, leave y numeric and do stratification by quantiles.
-    if (exists('objective', where = params) &&
-        is.character(params$objective)) {
-      # If 'objective' provided in params, assume that y is a classification label
-      # unless objective is reg:squarederror
-      if (params$objective != 'reg:squarederror')
-        y <- factor(y)
+    if (is.character(objective)) {
+      y <- convert.labels(y, params$objective)
     } else {
       # If no 'objective' given in params, it means that user either wants to
       # use the default 'reg:squarederror' objective or has provided a custom
       # obj function.  Here, assume classification setting when y has 5 or less
       # unique values:
-      if (length(unique(y)) <= 5)
+      if (length(unique(y)) <= 5) {
         y <- factor(y)
+      }
     }
     folds <- xgb.createFolds(y, nfold)
   } else {
@@ -307,6 +320,68 @@ xgb.createFolds <- function(y, k = 10)
 #' @name xgboost-deprecated
 NULL
 
+#' Do not use \code{\link[base]{saveRDS}} or \code{\link[base]{save}} for long-term archival of
+#' models. Instead, use \code{\link{xgb.save}} or \code{\link{xgb.save.raw}}.
+#'
+#' It is a common practice to use the built-in \code{\link[base]{saveRDS}} function (or
+#' \code{\link[base]{save}}) to persist R objects to the disk. While it is possible to persist
+#' \code{xgb.Booster} objects using \code{\link[base]{saveRDS}}, it is not advisable to do so if
+#' the model is to be accessed in the future. If you train a model with the current version of
+#' XGBoost and persist it with \code{\link[base]{saveRDS}}, the model is not guaranteed to be
+#' accessible in later releases of XGBoost. To ensure that your model can be accessed in future
+#' releases of XGBoost, use \code{\link{xgb.save}} or \code{\link{xgb.save.raw}} instead.
+#'
+#' @details
+#' Use \code{\link{xgb.save}} to save the XGBoost model as a stand-alone file. You may opt into
+#' the JSON format by specifying the JSON extension. To read the model back, use
+#' \code{\link{xgb.load}}.
+#'
+#' Use \code{\link{xgb.save.raw}} to save the XGBoost model as a sequence (vector) of raw bytes
+#' in a future-proof manner. Future releases of XGBoost will be able to read the raw bytes and
+#' re-construct the corresponding model. To read the model back, use \code{\link{xgb.load.raw}}.
+#' The \code{\link{xgb.save.raw}} function is useful if you'd like to persist the XGBoost model
+#' as part of another R object.
+#'
+#' Note: Do not use \code{\link{xgb.serialize}} to store models long-term. It persists not only the
+#' model but also internal configurations and parameters, and its format is not stable across
+#' multiple XGBoost versions. Use \code{\link{xgb.serialize}} only for checkpointing.
+#'
+#' For more details and explanation about model persistence and archival, consult the page
+#' \url{https://xgboost.readthedocs.io/en/latest/tutorials/saving_model.html}.
+#'
+#' @examples
+#' data(agaricus.train, package='xgboost')
+#' bst <- xgboost(data = agaricus.train$data, label = agaricus.train$label, max_depth = 2,
+#'                eta = 1, nthread = 2, nrounds = 2, objective = "binary:logistic")
+#'
+#' # Save as a stand-alone file; load it with xgb.load()
+#' xgb.save(bst, 'xgb.model')
+#' bst2 <- xgb.load('xgb.model')
+#'
+#' # Save as a stand-alone file (JSON); load it with xgb.load()
+#' xgb.save(bst, 'xgb.model.json')
+#' bst2 <- xgb.load('xgb.model.json')
+#' if (file.exists('xgb.model.json')) file.remove('xgb.model.json')
+#'
+#' # Save as a raw byte vector; load it with xgb.load.raw()
+#' xgb_bytes <- xgb.save.raw(bst)
+#' bst2 <- xgb.load.raw(xgb_bytes)
+#'
+#' # Persist XGBoost model as part of another R object
+#' obj <- list(xgb_model_bytes = xgb.save.raw(bst), description = "My first XGBoost model")
+#' # Persist the R object. Here, saveRDS() is okay, since it doesn't persist
+#' # xgb.Booster directly. What's being persisted is the future-proof byte representation
+#' # as given by xgb.save.raw().
+#' saveRDS(obj, 'my_object.rds')
+#' # Read back the R object
+#' obj2 <- readRDS('my_object.rds')
+#' # Re-construct xgb.Booster object from the bytes
+#' bst2 <- xgb.load.raw(obj2$xgb_model_bytes)
+#' if (file.exists('my_object.rds')) file.remove('my_object.rds')
+#'
+#' @name a-compatibility-note-for-saveRDS-save
+NULL
+
 # Lookup table for the deprecated parameters bookkeeping
 depr_par_lut <- matrix(c(
   'print.every.n', 'print_every_n',
@@ -315,8 +390,8 @@ depr_par_lut <- matrix(c(
   'with.stats', 'with_stats',
   'numberOfClusters', 'n_clusters',
   'features.keep', 'features_keep',
-  'plot.height','plot_height',
-  'plot.width','plot_width',
+  'plot.height', 'plot_height',
+  'plot.width', 'plot_width',
   'n_first_tree', 'trees',
   'dummy', 'DUMMY'
 ), ncol = 2, byrow = TRUE)
@@ -329,20 +404,20 @@ colnames(depr_par_lut) <- c('old', 'new')
 check.deprecation <- function(..., env = parent.frame()) {
   pars <- list(...)
   # exact and partial matches
-  all_match <- pmatch(names(pars), depr_par_lut[,1])
+  all_match <- pmatch(names(pars), depr_par_lut[, 1])
   # indices of matched pars' names
   idx_pars <- which(!is.na(all_match))
   if (length(idx_pars) == 0) return()
   # indices of matched LUT rows
   idx_lut <- all_match[idx_pars]
   # which of idx_lut were the exact matches?
-  ex_match <- depr_par_lut[idx_lut,1] %in% names(pars)
+  ex_match <- depr_par_lut[idx_lut, 1] %in% names(pars)
   for (i in seq_along(idx_pars)) {
     pars_par <- names(pars)[idx_pars[i]]
     old_par <- depr_par_lut[idx_lut[i], 1]
     new_par <- depr_par_lut[idx_lut[i], 2]
     if (!ex_match[i]) {
-      warning("'", pars_par, "' was partially matched to '", old_par,"'")
+      warning("'", pars_par, "' was partially matched to '", old_par, "'")
     }
     .Deprecated(new_par, old = old_par, package = 'xgboost')
     if (new_par != 'NULL') {
