@@ -70,11 +70,14 @@ class TreeUpdater : public Configurable {
    *         the prediction cache. If true, the prediction cache will have been
    *         updated by the time this function returns.
    */
-  virtual bool UpdatePredictionCache(const DMatrix* data,
-                                     HostDeviceVector<bst_float>* out_preds) {
-    // Remove unused parameter compiler warning.
-    (void) data;
-    (void) out_preds;
+  virtual bool UpdatePredictionCache(const DMatrix* /*data*/,
+                                     HostDeviceVector<bst_float>* /*out_preds*/) {
+    return false;
+  }
+
+  virtual bool UpdatePredictionCacheMulticlass(const DMatrix* /*data*/,
+                                               HostDeviceVector<bst_float>* /*out_preds*/,
+                                               const int /*gid*/, const int /*ngroup*/) {
     return false;
   }
 
