@@ -1,27 +1,46 @@
 /*!
- * Copyright 2019 by Contributors
+ * Copyright 2019-2020 by Contributors
  */
+#include <utility>
+
 #include "xgboost/json.h"
 #include "xgboost/logging.h"
 #include "gbtree_model.h"
+#include "gbtree.h"
 
 namespace xgboost {
 namespace gbm {
-
 void GBTreeModel::Save(dmlc::Stream* fo) const {
   CHECK_EQ(param.num_trees, static_cast<int32_t>(trees.size()));
-  fo->Write(&param, sizeof(param));
+
+  if (DMLC_IO_NO_ENDIAN_SWAP) {
+    fo->Write(&param, sizeof(param));
+  } else {
+    auto x = param.ByteSwap();
+    fo->Write(&x, sizeof(x));
+  }
   for (const auto & tree : trees) {
     tree->Save(fo);
   }
   if (tree_info.size() != 0) {
-    fo->Write(dmlc::BeginPtr(tree_info), sizeof(int32_t) * tree_info.size());
+    if (DMLC_IO_NO_ENDIAN_SWAP) {
+      fo->Write(dmlc::BeginPtr(tree_info), sizeof(int32_t) * tree_info.size());
+    } else {
+      for (const auto& e : tree_info) {
+        auto x = e;
+        dmlc::ByteSwap(&x, sizeof(x), 1);
+        fo->Write(&x, sizeof(x));
+      }
+    }
   }
 }
 
 void GBTreeModel::Load(dmlc::Stream* fi) {
   CHECK_EQ(fi->Read(&param, sizeof(param)), sizeof(param))
       << "GBTree: invalid model file";
+  if (!DMLC_IO_NO_ENDIAN_SWAP) {
+    param = param.ByteSwap();
+  }
   trees.clear();
   trees_to_update.clear();
   for (int32_t i = 0; i < param.num_trees; ++i) {
@@ -31,9 +50,16 @@ void GBTreeModel::Load(dmlc::Stream* fi) {
   }
   tree_info.resize(param.num_trees);
   if (param.num_trees != 0) {
-    CHECK_EQ(
-        fi->Read(dmlc::BeginPtr(tree_info), sizeof(int32_t) * param.num_trees),
-        sizeof(int32_t) * param.num_trees);
+    if (DMLC_IO_NO_ENDIAN_SWAP) {
+      CHECK_EQ(
+          fi->Read(dmlc::BeginPtr(tree_info), sizeof(int32_t) * param.num_trees),
+          sizeof(int32_t) * param.num_trees);
+    } else {
+      for (auto& info : tree_info) {
+        CHECK_EQ(fi->Read(&info, sizeof(int32_t)), sizeof(int32_t));
+        dmlc::ByteSwap(&info, sizeof(info), 1);
+      }
+    }
   }
 }
 
@@ -41,15 +67,14 @@ void GBTreeModel::SaveModel(Json* p_out) const {
   auto& out = *p_out;
   CHECK_EQ(param.num_trees, static_cast<int>(trees.size()));
   out["gbtree_model_param"] = ToJson(param);
-  std::vector<Json> trees_json;
-  size_t t = 0;
-  for (auto const& tree : trees) {
+  std::vector<Json> trees_json(trees.size());
+
+  for (size_t t = 0; t < trees.size(); ++t) {
+    auto const& tree = trees[t];
     Json tree_json{Object()};
     tree->SaveModel(&tree_json);
-    // The field is not used in XGBoost, but might be useful for external project.
-    tree_json["id"] = Integer(t);
-    trees_json.emplace_back(tree_json);
-    t++;
+    tree_json["id"] = Integer(static_cast<Integer::Int>(t));
+    trees_json[t] = std::move(tree_json);
   }
 
   std::vector<Json> tree_info_json(tree_info.size());
@@ -70,9 +95,10 @@ void GBTreeModel::LoadModel(Json const& in) {
   auto const& trees_json = get<Array const>(in["trees"]);
   trees.resize(trees_json.size());
 
-  for (size_t t = 0; t < trees.size(); ++t) {
-    trees[t].reset( new RegTree() );
-    trees[t]->LoadModel(trees_json[t]);
+  for (size_t t = 0; t < trees_json.size(); ++t) {  // NOLINT
+    auto tree_id = get<Integer>(trees_json[t]["id"]);
+    trees.at(tree_id).reset(new RegTree());
+    trees.at(tree_id)->LoadModel(trees_json[t]);
   }
 
   tree_info.resize(param.num_trees);
