@@ -29,6 +29,7 @@
 
 #include "xgboost/host_device_vector.h"
 #include "../common/math.h"
+#include "../common/threading_utils.h"
 #include "metric_common.h"
 
 namespace {
@@ -111,10 +112,9 @@ struct EvalAMS : public Metric {
     PredIndPairContainer rec(ndata);
 
     const auto &h_preds = preds.ConstHostVector();
-    #pragma omp parallel for schedule(static)
-    for (bst_omp_uint i = 0; i < ndata; ++i) {
+    common::ParallelFor(ndata, [&](bst_omp_uint i) {
       rec[i] = std::make_pair(h_preds[i], i);
-    }
+    });
     XGBOOST_PARALLEL_SORT(rec.begin(), rec.end(), common::CmpFirst);
     auto ntop = static_cast<unsigned>(ratio_ * ndata);
     if (ntop == 0) ntop = ndata;
@@ -156,126 +156,6 @@ struct EvalAMS : public Metric {
   float ratio_;
 };
 
-/*! \brief Area Under Curve, for both classification and rank computed on CPU */
-struct EvalAuc : public Metric {
- private:
-  // This is used to compute the AUC metrics on the GPU - for ranking tasks and
-  // for training jobs that run on the GPU.
-  std::unique_ptr<xgboost::Metric> auc_gpu_;
-
-  template <typename WeightPolicy>
-  bst_float Eval(const HostDeviceVector<bst_float> &preds,
-                 const MetaInfo &info,
-                 bool distributed,
-                 const std::vector<unsigned> &gptr) {
-    const auto ngroups = static_cast<bst_omp_uint>(gptr.size() - 1);
-    // sum of all AUC's across all query groups
-    double sum_auc = 0.0;
-    int auc_error = 0;
-    const auto& labels = info.labels_.ConstHostVector();
-    const auto &h_preds = preds.ConstHostVector();
-
-    #pragma omp parallel reduction(+:sum_auc, auc_error) if (ngroups > 1)
-    {
-      // Each thread works on a distinct group and sorts the predictions in that group
-      PredIndPairContainer rec;
-      #pragma omp for schedule(static)
-      for (bst_omp_uint group_id = 0; group_id < ngroups; ++group_id) {
-        // Same thread can work on multiple groups one after another; hence, resize
-        // the predictions array based on the current group
-        rec.resize(gptr[group_id + 1] - gptr[group_id]);
-        #pragma omp parallel for schedule(static) if (!omp_in_parallel())
-        for (bst_omp_uint j = gptr[group_id]; j < gptr[group_id + 1]; ++j) {
-          rec[j - gptr[group_id]] = {h_preds[j], j};
-        }
-
-        XGBOOST_PARALLEL_SORT(rec.begin(), rec.end(), common::CmpFirst);
-        // calculate AUC
-        double sum_pospair = 0.0;
-        double sum_npos = 0.0, sum_nneg = 0.0, buf_pos = 0.0, buf_neg = 0.0;
-        for (size_t j = 0; j < rec.size(); ++j) {
-          const bst_float wt = WeightPolicy::GetWeightOfSortedRecord(info, rec, j, group_id);
-          const bst_float ctr = labels[rec[j].second];
-          // keep bucketing predictions in same bucket
-          if (j != 0 && rec[j].first != rec[j - 1].first) {
-            sum_pospair += buf_neg * (sum_npos + buf_pos * 0.5);
-            sum_npos += buf_pos;
-            sum_nneg += buf_neg;
-            buf_neg = buf_pos = 0.0f;
-          }
-          buf_pos += ctr * wt;
-          buf_neg += (1.0f - ctr) * wt;
-        }
-        sum_pospair += buf_neg * (sum_npos + buf_pos * 0.5);
-        sum_npos += buf_pos;
-        sum_nneg += buf_neg;
-        // check weird conditions
-        if (sum_npos <= 0.0 || sum_nneg <= 0.0) {
-          auc_error += 1;
-        } else {
-          // this is the AUC
-          sum_auc += sum_pospair / (sum_npos * sum_nneg);
-        }
-      }
-    }
-
-    // Report average AUC across all groups
-    // In distributed mode, workers which only contains pos or neg samples
-    // will be ignored when aggregate AUC.
-    bst_float dat[2] = {0.0f, 0.0f};
-    if (auc_error < static_cast<int>(ngroups)) {
-      dat[0] = static_cast<bst_float>(sum_auc);
-      dat[1] = static_cast<bst_float>(static_cast<int>(ngroups) - auc_error);
-    }
-    if (distributed) {
-      rabit::Allreduce<rabit::op::Sum>(dat, 2);
-    }
-    CHECK_GT(dat[1], 0.0f)
-      << "AUC: the dataset only contains pos or neg samples";
-    return dat[0] / dat[1];
-  }
-
- public:
-  bst_float Eval(const HostDeviceVector<bst_float> &preds,
-                 const MetaInfo &info,
-                 bool distributed) override {
-    CHECK_NE(info.labels_.Size(), 0U) << "label set cannot be empty";
-    CHECK_EQ(preds.Size(), info.labels_.Size())
-        << "label size predict size not match";
-    std::vector<unsigned> tgptr(2, 0);
-    tgptr[1] = static_cast<unsigned>(info.labels_.Size());
-
-    const auto &gptr = info.group_ptr_.empty() ? tgptr : info.group_ptr_;
-    CHECK_EQ(gptr.back(), info.labels_.Size())
-        << "EvalAuc: group structure must match number of prediction";
-
-    // For ranking task, weights are per-group
-    // For binary classification task, weights are per-instance
-    const bool is_ranking_task =
-      !info.group_ptr_.empty() && info.weights_.Size() != info.num_row_;
-
-    // Check if we have a GPU assignment; else, revert back to CPU
-    if (tparam_->gpu_id >= 0) {
-      if (!auc_gpu_) {
-        // Check and see if we have the GPU metric registered in the internal registry
-        auc_gpu_.reset(GPUMetric::CreateGPUMetric(this->Name(), tparam_));
-      }
-
-      if (auc_gpu_) {
-        return auc_gpu_->Eval(preds, info, distributed);
-      }
-    }
-
-    if (is_ranking_task) {
-      return Eval<PerGroupWeightPolicy>(preds, info, distributed, gptr);
-    } else {
-      return Eval<PerInstanceWeightPolicy>(preds, info, distributed, gptr);
-    }
-  }
-
-  const char *Name() const override { return "auc"; }
-};
-
 /*! \brief Evaluate rank list */
 struct EvalRank : public Metric, public EvalRankConfig {
  private:
@@ -316,19 +196,25 @@ struct EvalRank : public Metric, public EvalRankConfig {
       const auto &labels = info.labels_.ConstHostVector();
       const auto &h_preds = preds.ConstHostVector();
 
+      dmlc::OMPException exc;
       #pragma omp parallel reduction(+:sum_metric)
       {
-        // each thread takes a local rec
-        PredIndPairContainer rec;
-        #pragma omp for schedule(static)
-        for (bst_omp_uint k = 0; k < ngroups; ++k) {
-          rec.clear();
-          for (unsigned j = gptr[k]; j < gptr[k + 1]; ++j) {
-            rec.emplace_back(h_preds[j], static_cast<int>(labels[j]));
+        exc.Run([&]() {
+          // each thread takes a local rec
+          PredIndPairContainer rec;
+          #pragma omp for schedule(static)
+          for (bst_omp_uint k = 0; k < ngroups; ++k) {
+            exc.Run([&]() {
+              rec.clear();
+              for (unsigned j = gptr[k]; j < gptr[k + 1]; ++j) {
+                rec.emplace_back(h_preds[j], static_cast<int>(labels[j]));
+              }
+              sum_metric += this->EvalGroup(&rec);
+            });
           }
-          sum_metric += this->EvalGroup(&rec);
-        }
+        });
       }
+      exc.Rethrow();
     }
 
     if (distributed) {
@@ -526,66 +412,75 @@ struct EvalAucPR : public Metric {
     const auto &h_labels = info.labels_.ConstHostVector();
     const auto &h_preds = preds.ConstHostVector();
 
+    dmlc::OMPException exc;
     #pragma omp parallel reduction(+:sum_auc, auc_error) if (ngroups > 1)
     {
-      // Each thread works on a distinct group and sorts the predictions in that group
-      PredIndPairContainer rec;
-      #pragma omp for schedule(static)
-      for (bst_omp_uint group_id = 0; group_id < ngroups; ++group_id) {
-        double total_pos = 0.0;
-        double total_neg = 0.0;
-        // Same thread can work on multiple groups one after another; hence, resize
-        // the predictions array based on the current group
-        rec.resize(gptr[group_id + 1] - gptr[group_id]);
-        #pragma omp parallel for schedule(static) reduction(+:total_pos, total_neg) \
-          if (!omp_in_parallel())  // NOLINT
-        for (bst_omp_uint j = gptr[group_id]; j < gptr[group_id + 1]; ++j) {
-          const bst_float wt = WeightPolicy::GetWeightOfInstance(info, j, group_id);
-          total_pos += wt * h_labels[j];
-          total_neg += wt * (1.0f - h_labels[j]);
-          rec[j - gptr[group_id]] = {h_preds[j], j};
-        }
-
-        // we need pos > 0 && neg > 0
-        if (total_pos <= 0.0 || total_neg <= 0.0) {
-          auc_error += 1;
-          continue;
-        }
-
-        XGBOOST_PARALLEL_SORT(rec.begin(), rec.end(), common::CmpFirst);
-
-        // calculate AUC
-        double tp = 0.0, prevtp = 0.0, fp = 0.0, prevfp = 0.0, h = 0.0, a = 0.0, b = 0.0;
-        for (size_t j = 0; j < rec.size(); ++j) {
-          const bst_float wt = WeightPolicy::GetWeightOfSortedRecord(info, rec, j, group_id);
-          tp += wt * h_labels[rec[j].second];
-          fp += wt * (1.0f - h_labels[rec[j].second]);
-          if ((j < rec.size() - 1 && rec[j].first != rec[j + 1].first) || j  == rec.size() - 1) {
-            if (tp == prevtp) {
-              a = 1.0;
-              b = 0.0;
-            } else {
-              h = (fp - prevfp) / (tp - prevtp);
-              a = 1.0 + h;
-              b = (prevfp - h * prevtp) / total_pos;
+      exc.Run([&]() {
+        // Each thread works on a distinct group and sorts the predictions in that group
+        PredIndPairContainer rec;
+        #pragma omp for schedule(static)
+        for (bst_omp_uint group_id = 0; group_id < ngroups; ++group_id) {
+          exc.Run([&]() {
+            double total_pos = 0.0;
+            double total_neg = 0.0;
+            // Same thread can work on multiple groups one after another; hence, resize
+            // the predictions array based on the current group
+            rec.resize(gptr[group_id + 1] - gptr[group_id]);
+            #pragma omp parallel for schedule(static) reduction(+:total_pos, total_neg) \
+              if (!omp_in_parallel())  // NOLINT
+            for (bst_omp_uint j = gptr[group_id]; j < gptr[group_id + 1]; ++j) {
+              exc.Run([&]() {
+                const bst_float wt = WeightPolicy::GetWeightOfInstance(info, j, group_id);
+                total_pos += wt * h_labels[j];
+                total_neg += wt * (1.0f - h_labels[j]);
+                rec[j - gptr[group_id]] = {h_preds[j], j};
+              });
             }
-            if (0.0 != b) {
-              sum_auc += (tp / total_pos - prevtp / total_pos -
-                          b / a * (std::log(a * tp / total_pos + b) -
-                                   std::log(a * prevtp / total_pos + b))) / a;
-            } else {
-              sum_auc += (tp / total_pos - prevtp / total_pos) / a;
+
+            // we need pos > 0 && neg > 0
+            if (total_pos <= 0.0 || total_neg <= 0.0) {
+              auc_error += 1;
+              return;
             }
-            prevtp = tp;
-            prevfp = fp;
-          }
+
+            XGBOOST_PARALLEL_SORT(rec.begin(), rec.end(), common::CmpFirst);
+
+            // calculate AUC
+            double tp = 0.0, prevtp = 0.0, fp = 0.0, prevfp = 0.0, h = 0.0, a = 0.0, b = 0.0;
+            for (size_t j = 0; j < rec.size(); ++j) {
+              const bst_float wt = WeightPolicy::GetWeightOfSortedRecord(info, rec, j, group_id);
+              tp += wt * h_labels[rec[j].second];
+              fp += wt * (1.0f - h_labels[rec[j].second]);
+              if ((j < rec.size() - 1 && rec[j].first != rec[j + 1].first) ||
+                  j  == rec.size() - 1) {
+                if (tp == prevtp) {
+                  a = 1.0;
+                  b = 0.0;
+                } else {
+                  h = (fp - prevfp) / (tp - prevtp);
+                  a = 1.0 + h;
+                  b = (prevfp - h * prevtp) / total_pos;
+                }
+                if (0.0 != b) {
+                  sum_auc += (tp / total_pos - prevtp / total_pos -
+                              b / a * (std::log(a * tp / total_pos + b) -
+                                      std::log(a * prevtp / total_pos + b))) / a;
+                } else {
+                  sum_auc += (tp / total_pos - prevtp / total_pos) / a;
+                }
+                prevtp = tp;
+                prevfp = fp;
+              }
+            }
+            // sanity check
+            if (tp < 0 || prevtp < 0 || fp < 0 || prevfp < 0) {
+              CHECK(!auc_error) << "AUC-PR: error in calculation";
+            }
+          });
         }
-        // sanity check
-        if (tp < 0 || prevtp < 0 || fp < 0 || prevfp < 0) {
-          CHECK(!auc_error) << "AUC-PR: error in calculation";
-        }
-      }
+      });
     }
+    exc.Rethrow();
 
     // Report average AUC-PR across all groups
     // In distributed mode, workers which only contains pos or neg samples
@@ -649,10 +544,6 @@ XGBOOST_REGISTER_METRIC(AMS, "ams")
 .describe("AMS metric for higgs.")
 .set_body([](const char* param) { return new EvalAMS(param); });
 
-XGBOOST_REGISTER_METRIC(Auc, "auc")
-.describe("Area under curve for both classification and rank.")
-.set_body([](const char*) { return new EvalAuc(); });
-
 XGBOOST_REGISTER_METRIC(AucPR, "aucpr")
 .describe("Area under PR curve for both classification and rank.")
 .set_body([](const char*) { return new EvalAucPR(); });
