@@ -77,8 +77,9 @@ class ColMaker: public TreeUpdater {
     if (column_densities_.empty()) {
       std::vector<size_t> column_size(dmat->Info().num_col_);
       for (const auto &batch : dmat->GetBatches<SortedCSCPage>()) {
+        auto page = batch.GetView();
         for (auto i = 0u; i < batch.Size(); i++) {
-          column_size[i] += batch[i].size();
+          column_size[i] += page[i].size();
         }
       }
       column_densities_.resize(column_size.size());
@@ -263,12 +264,16 @@ class ColMaker: public TreeUpdater {
       const MetaInfo& info = fmat.Info();
       // setup position
       const auto ndata = static_cast<bst_omp_uint>(info.num_row_);
+      dmlc::OMPException exc;
       #pragma omp parallel for schedule(static)
       for (bst_omp_uint ridx = 0; ridx < ndata; ++ridx) {
-        const int tid = omp_get_thread_num();
-        if (position_[ridx] < 0) continue;
-        stemp_[tid][position_[ridx]].stats.Add(gpair[ridx]);
+        exc.Run([&]() {
+          const int tid = omp_get_thread_num();
+          if (position_[ridx] < 0) return;
+          stemp_[tid][position_[ridx]].stats.Add(gpair[ridx]);
+        });
       }
+      exc.Rethrow();
       // sum the per thread statistics together
       for (int nid : qexpand) {
         GradStats stats;
@@ -446,29 +451,30 @@ class ColMaker: public TreeUpdater {
           std::max(static_cast<int>(num_features / this->nthread_ / 32), 1);
 #endif  // defined(_OPENMP)
       {
-        dmlc::OMPException omp_handler;
+        auto page = batch.GetView();
+        dmlc::OMPException exc;
 #pragma omp parallel for schedule(dynamic, batch_size)
         for (bst_omp_uint i = 0; i < num_features; ++i) {
-          omp_handler.Run([&]() {
+          exc.Run([&]() {
             auto evaluator = tree_evaluator_.GetEvaluator();
             bst_feature_t const fid = feat_set[i];
             int32_t const tid = omp_get_thread_num();
-            auto c = batch[fid];
+            auto c = page[fid];
             const bool ind =
                 c.size() != 0 && c[0].fvalue == c[c.size() - 1].fvalue;
             if (colmaker_train_param_.NeedForwardSearch(
                     param_.default_direction, column_densities_[fid], ind)) {
               this->EnumerateSplit(c.data(), c.data() + c.size(), +1, fid,
-                                   gpair, stemp_[tid], evaluator);
+                                  gpair, stemp_[tid], evaluator);
             }
             if (colmaker_train_param_.NeedBackwardSearch(
                     param_.default_direction)) {
               this->EnumerateSplit(c.data() + c.size() - 1, c.data() - 1, -1,
-                                   fid, gpair, stemp_[tid], evaluator);
+                                  fid, gpair, stemp_[tid], evaluator);
             }
           });
         }
-        omp_handler.Rethrow();
+        exc.Rethrow();
       }
     }
     // find splits at current level, do split per level
@@ -519,8 +525,7 @@ class ColMaker: public TreeUpdater {
       // so that they are ignored in future statistics collection
       const auto ndata = static_cast<bst_omp_uint>(p_fmat->Info().num_row_);
 
-#pragma omp parallel for schedule(static)
-      for (bst_omp_uint ridx = 0; ridx < ndata; ++ridx) {
+      common::ParallelFor(ndata, [&](bst_omp_uint ridx) {
         CHECK_LT(ridx, position_.size())
             << "ridx exceed bound " << "ridx="<<  ridx << " pos=" << position_.size();
         const int nid = this->DecodePosition(ridx);
@@ -537,7 +542,7 @@ class ColMaker: public TreeUpdater {
             this->SetEncodePosition(ridx, tree[nid].RightChild());
           }
         }
-      }
+      });
     }
     // customization part
     // synchronize the best solution of each node
@@ -562,11 +567,11 @@ class ColMaker: public TreeUpdater {
       std::sort(fsplits.begin(), fsplits.end());
       fsplits.resize(std::unique(fsplits.begin(), fsplits.end()) - fsplits.begin());
       for (const auto &batch : p_fmat->GetBatches<SortedCSCPage>()) {
+        auto page = batch.GetView();
         for (auto fid : fsplits) {
-          auto col = batch[fid];
+          auto col = page[fid];
           const auto ndata = static_cast<bst_omp_uint>(col.size());
-#pragma omp parallel for schedule(static)
-          for (bst_omp_uint j = 0; j < ndata; ++j) {
+          common::ParallelFor(ndata, [&](bst_omp_uint j) {
             const bst_uint ridx = col[j].index;
             const int nid = this->DecodePosition(ridx);
             const bst_float fvalue = col[j].fvalue;
@@ -578,7 +583,7 @@ class ColMaker: public TreeUpdater {
                 this->SetEncodePosition(ridx, tree[nid].RightChild());
               }
             }
-          }
+          });
         }
       }
     }
